/* TEDSETUP.C
**
** Copyright (c)2002-2004 Andrew Durdin. (andy@durdin.net)
** This software is provided 'as-is', without any express or implied warranty.
** In no event will the authors be held liable for any damages arising from
** the use of this software.
** Permission is granted to anyone to use this software for any purpose, including
** commercial applications, and to alter it and redistribute it freely, subject
** to the following restrictions:
**    1. The origin of this software must not be misrepresented; you must not
**       claim that you wrote the original software. If you use this software in
**       a product, an acknowledgment in the product documentation would be
**       appreciated but is not required.
**    2. Altered source versions must be plainly marked as such, and must not be
**       misrepresented as being the original software.
**    3. This notice may not be removed or altered from any source distribution.
*/

#include <string.h>
#include <stdio.h>
#include <dir.h>

typedef unsigned short int word;

typedef struct {
	word episode;
	word version;
	long	imglen;
	long	maphead;
	long	egahead;
	long	egadict;
	long	audhead;
	long	auddict;
	long	tiledat;
	word	tilechk;
	word numbg;
	word numfg;
//	word numchunks;
} KEEN_EXE_INFO;

#define NUM_VERSIONS	9
KEEN_EXE_INFO exe_info[NUM_VERSIONS] = {
//   ep, ver, image,   maphead, egahead, egadict, audhead, auddict, tiledat, tch, nbg,  nfg
	{ 4, 100, 0x3C210, 0x235B0, 0x1FE10, 0x343F6, 0x1FB80, 0x33FF6, 0x23742, 523, 1296, 2916 },
	{ 4, 110, 0x3C4A0, 0x23780, 0x1FFE0, 0x34682, 0x1FD50, 0x34282, 0x23912, 523, 1296, 2916 },
	{ 4, 120, 0x3C750, 0x239E0, 0x20230, 0x34934, 0x1FFA0, 0x34534, 0x23B72, 523, 1296, 2916 },
	{ 4, 140, 0x3D740, 0x24830, 0x21080, 0x358F6, 0x20DF0, 0x354F6, 0x249C2, 523, 1296, 2916 },
	{ 5, 100, 0x3D220, 0x24980, 0x20FB0, 0x3518A, 0x20C70, 0x34D8A, 0x24B12, 450, 1512, 2952 },
	{ 5, 140, 0x3E370, 0x25990, 0x21FC0, 0x362C4, 0x21C80, 0x35EC4, 0x25B22, 450, 1512, 2952 },
	{ 6, 100, 0x3E330, 0x23C50, 0x1FB20, 0x35D12, 0x1F820, 0x35912, 0x23DE2, 438, 2376, 2736 },
	{ 6, 140, 0x3F950, 0x25080, 0x20F50, 0x372EE, 0x20C50, 0x36EEE, 0x25212, 438, 2376, 2736 },
	{ 6, 999, 0x37650, 0x1F560, 0x1C230, 0x2FA40, 0x1BFE0, 0x2F630, 0x1F6F2, 324, 1296, 2736 }
// Lengths:            0x192   to maphed  1024   to egahed  1024    2nb+7nf
};

typedef struct {
	word maptype;	//bit 0=bkgnd/1=frgnd/2=info
	word tsize;		//1=8/2=16/3=32

	word numtplanes,oldtilenum;
	long tileinfooff[10];
	word tileinfolen[10];
	char tnames[10][8];

	word numtmplanes,oldtilemnum;
	long tileinfomoff[10];
	word tileinfomlen[10];
	char tmnames[10][8];


	word RLEWtag;
	long dataoffsets[100];
	long datalengths[100];

	word NumIconRows;
} MAPTHEAD_STRUCT;


typedef struct {
	word RLEWtag ;
	long dataoffsets[100];
} MAPHEAD_STRUCT;

typedef struct {
	word num8,num8m,num16,num16m,num32,num32m ;
	word off8,off8m,off16,off16m,off32,off32m ;
	word numpics,numpicm,numsprites ;
	word offpic,offpicm,offsprites ;
	word offpicstr,offpicmstr,offsprstr ;
	word numexterns,offexterns ;
} GFXINFOE_STRUCT;

typedef struct {
	word mzid ;
	word image_l ;
	word image_h ;
	word num_relocs ;
	word header_size ;
	word min_paras ;
	word max_paras ;
	word init_ss ;
	word init_sp ;
	word checksum ;
	word init_ip ;
	word init_cs ;
	word reloc_offset ;
	word overlay_num ;
} EXE_HEADER;

#define EXEMZ	0x5A4D
#define EXEZM	0x4D5A

// Returns the size of the loaded image of the program, or -1 on error
void get_exe_image_size( FILE *f, long *imglen, long *headerlen ) {
	EXE_HEADER head;

	rewind( f );
	// Read the header from the file if we can
	if( fread( &head, sizeof( EXE_HEADER ), 1, f ) == 1 ) {
		// Check that the 'MZ' id is present
		if( head.mzid == EXEMZ || head.mzid == EXEZM ) {
			// Calculate the image size
			*imglen = ((long)head.image_h - 1) * 512L + head.image_l - (long)head.header_size * 16L;
			*headerlen = (long)head.header_size * 16L;

			printf( "Image size: %ld or 0x%lX, Header size %ld or 0x%lX\n",
				*imglen, *imglen, *headerlen, *headerlen );

			return;
		} else {
			printf( "Error: invalid EXE!\n" );
		}
	} else {
		printf( "Error reading EXE header!\n" );
	}

	// If we got here, something failed
	*imglen = *headerlen = 0;
}

#define READBUF_SIZE	1024
static unsigned char readbuf[ READBUF_SIZE ];
int extract( FILE *f, char *fname, long offset, long size ) {
	FILE *fo;
	int ret;

//	printf( "Extracting '%s' size %lX from %lX!\n", fname, size, offset );

	fseek( f, offset, SEEK_SET );
	fo = fopen( fname, "wb" );
	if( fo ) {
		while( size > READBUF_SIZE ) {
			fread( readbuf, READBUF_SIZE, 1, f );
			fwrite( readbuf, READBUF_SIZE, 1, fo );
			size -= READBUF_SIZE;
		}
		if( size > 0 ) {
			fread( readbuf, size, 1, f );
			fwrite( readbuf, size, 1, fo );
		}
		fclose( fo );
		ret = 0;
	} else {
		printf( "Error: can't create file '%s'!\n", fname );
		ret = -1;
	}

	return ret;
}

typedef struct {
	long offset[3];
	word len[3];
	word width;
	word height;
	char name[16];
} MAP_HEADER;

#define NEARTAG 0xA7
#define FARTAG 0xA8

void decarmack( word far *source, word far *dest, long inlen, long outlen ) {
	word ch, chhigh, count, offset;
	word far *copyptr, far *inptr, far *outptr;

	word maxoffset = outlen / 2;

	if( !source || !dest ) {
		printf(	"Error: NULL input to decarmack()\n" );
		return;
	}

	/* Let's have the output length in words */
	outlen /= 2;

	inptr = source;
	outptr = dest;

	while( outlen > 0 )
	{
		/* Read the current word */
		ch = *inptr++;
		inlen -= 2;

		/* Check the high byte */
		chhigh = (ch >> 8) & 0xFF;
		if( chhigh == NEARTAG )
		{
			/* The count is the low byte */
			count = ch & 0xFF;
			if( !count )
			{
				// have to insert a word containing the tag byte
				ch |= *((unsigned char far *)inptr)++;
				inlen--;
				*outptr++ = ch;
				outlen--;
			}
			else
			{
				offset = *((unsigned char far *)inptr)++;
				inlen--;
				copyptr = outptr - offset;

				outlen -= count;
				while( count-- )
					*outptr++ = *copyptr++;
			}
		}
		else if( chhigh == FARTAG )
		{
			count = ch & 0xFF;
			if( !count )
			{
				/* have to insert a word containing the tag byte */
				ch |= *((unsigned char far *)inptr)++;
				inlen--;
				*outptr++ = ch;
				outlen--;
			}
			else
			{
				offset = *inptr++;
				inlen -= 2;
				copyptr = dest + offset;
				if( offset > maxoffset ) {
					printf( "Error: Offset of %X exceeds max of %X!\n",
							offset, maxoffset );
				}
				outlen -= count;
				while( count-- )
					*outptr++ = *copyptr++;
			}
		}
		else
		{
			*outptr++ = ch;
			outlen--;
		}
	}
}

#define LEVELBUF_LEN	65500
static word far levelcbuf[ (LEVELBUF_LEN/2) ];
static word far levelubuf[ (LEVELBUF_LEN/2) ];

long do_level( long offset, FILE *fi, FILE *fo ) {
	MAP_HEADER head;
	int i;
	long totlen = 0;
	long headoffset = ftell( fo );
	long off = headoffset + sizeof( MAP_HEADER );
	word uncmplen;
	word *p;

	// First thing we do is get the header
	fseek( fi, offset, SEEK_SET );
	fread( &head, sizeof( MAP_HEADER ), 1, fi );

	// Leave space for it here
	fwrite( &head, sizeof( MAP_HEADER ), 1, fo );

	for( i = 0; i < 3; i++ ) {
		// Seek to the start of the plane and find out the uncompressed size
		fseek( fi, head.offset[i], SEEK_SET );
		fread( &uncmplen, sizeof( word ), 1, fi );

		// printf( "Decarmacking plane %d, length %X->%X from %ld\n", i, head.len[i], uncmplen, head.offset[i] );

		if( head.len[i] - 4 > LEVELBUF_LEN || uncmplen > LEVELBUF_LEN ) {
			printf( "Error: Plane %d is too large!\n", i );
			break;
		}

		// Read the compressed data
		fread( levelcbuf, head.len[i] - 2, 1, fi );

		// Decarmack the data into the buffer
		decarmack( levelcbuf, levelubuf, head.len[i] - 2, uncmplen );

		// Write the uncompressed data
		fwrite( levelubuf, uncmplen, 1, fo );

		// Update the header
		head.offset[i] = off;
		head.len[i] = (word)uncmplen;
		off += uncmplen;
		totlen += uncmplen;

	}

	// Write the header
	off = ftell( fo );
	fseek( fo, headoffset, SEEK_SET );
	fwrite( &head, sizeof( MAP_HEADER ), 1, fo );
	fseek( fo, off, SEEK_SET );
	totlen += sizeof( MAP_HEADER );

	return totlen;
}

MAPHEAD_STRUCT maphead;
MAPTHEAD_STRUCT mapthead;
GFXINFOE_STRUCT	gfxinfoe;

int do_maps( int v ) {
	FILE *fi, *fo;
	char fname[MAXPATH];
	long offset, len, id;
	int i;

	// Open the gamemaps file
	sprintf( fname, "gamemaps.ck%c", exe_info[v].episode + '0' );
	fi = fopen( fname, "rb" );
	if( !fi ) {
		printf( "Error: Couldn't open %s!\n", fname );
		return -1;
	}

	// Open the maptemp file
	sprintf( fname, "maptemp.ck%c", exe_info[v].episode + '0' );
	fo = fopen( fname, "wb" );
	if( !fo ) {
		printf( "Error: Couldn't open %s!\n", fname );
		fclose( fi );
		return -1;
	}

	// Now do the header
	id = 0x35444554;	// "TED5"
	fwrite( &id, sizeof( long ), 1, fo );
	id = 0x302E3176;	// "v1.0"
	fwrite( &id, sizeof( long ), 1, fo );

	// Now do all the maps
	offset = ftell( fo );
	for( i = 0; i < 100; i++ ) {
		if( maphead.dataoffsets[i] != 0 ) {
//			printf( "Decarmacking level %d... from %lX\n", i, maphead.dataoffsets[i] );
			len = do_level( maphead.dataoffsets[i], fi, fo );
			mapthead.dataoffsets[i] = offset;
			mapthead.datalengths[i] = len;
			offset += len;
		} else {
			mapthead.dataoffsets[i] = -1;
			mapthead.datalengths[i] = 0;
		}
	}
	return 0;
}

static int externalmaphead = 0;
static unsigned char mapheadname[32];



int extract_data( FILE *f, int v, long headerlen ) {
	FILE *fo;
	word w;
	int i;
	char fname[MAXPATH];
	char ckver = exe_info[v].episode + '0';

	// Verify that the file is OK by finding the RLE tag where we expect it
	fseek( f, headerlen + exe_info[v].maphead, SEEK_SET );
	fread( &w, sizeof( word ), 1, f );
	if( w != (word)0xABCD )
		return -1;

	// Now extract the egadict
	printf( "Extracting egadict...\n" );
	sprintf( fname, "egadict.ck%c", ckver );
	extract( f, fname, headerlen + exe_info[v].egadict, 1024 );

	// Now extract the egahead
	printf( "Extracting egahead...\n" );
	sprintf( fname, "egahead.ck%c", ckver );
	extract( f, fname, headerlen + exe_info[v].egahead, exe_info[v].maphead - exe_info[v].egahead );

	// Now we have to concoct the other data

	// Make the {b|p}atch files
	printf( "Creating batch file...\n" );
	sprintf( fname, "go%c.bat", ckver );
	fo = fopen( fname, "wt" );
	fprintf( fo, "@echo off\n"
/*				"if not exist fixmhead.exe goto just_run_it\n"
				"fixmhead ck%cmhead.obj maphead.ck%c\n\n"
				"on errorlevel 1 goto error\n" */
				"copy gamemaps.ck%c custmaps.ck%c\n"
				":just_run_it\n"
				"ck%cpatch patch%c.pat -nowait\n"
				":error\n", ckver, ckver, ckver, ckver, ckver, ckver );
	fclose( fo );

	sprintf( fname, "patch%c.pat", ckver );
	fo = fopen( fname, "wt" );
	fprintf( fo, "# Patch file for Keen - use with CK%cPATCH.EXE\n\n"
				"%%ext ck%c\n\n"
				"%%gamemaps custmaps.ck%c\n"
				"%%ckmhead.obj ckmhead.obj\n\n"
				"%%end\n", ckver, ckver, ckver, ckver );
	fclose( fo );

	// Make the "gfxinfoe"
	printf( "Creating gfxinfoe...\n" );
	sprintf( fname, "gfxinfoe.ck%c", ckver );
	fo = fopen( fname, "wb" );
	if( fo ) {
		memset( &gfxinfoe, 0, sizeof( GFXINFOE_STRUCT ) );
		gfxinfoe.num16 = exe_info[v].numbg;
		gfxinfoe.num16m = exe_info[v].numfg;
		gfxinfoe.off16 = exe_info[v].tilechk;
		gfxinfoe.off16m = exe_info[v].tilechk + exe_info[v].numbg;

		fwrite( &gfxinfoe, sizeof( GFXINFOE_STRUCT ), 1, fo );
		fclose( fo );
	} else {
		printf( "Error: can't create '%s'!\n", fname );
	}

	// Now we need to decarmack all the maps to make
	// "mapthead" and "maptemp"

	// First extract the maphead into our temporary buffer
	// foo tbd
	if(externalmaphead)
	{
		FILE *f2 = fopen( mapheadname, "rb" );
		fread( &maphead, sizeof( MAPHEAD_STRUCT ), 1, f2 );
		fclose( f2 );
	} else {
		fseek( f, headerlen + exe_info[v].maphead, SEEK_SET );
		fread( &maphead, sizeof( MAPHEAD_STRUCT ), 1, f );
	}

	// Initialise the fields of mapthead that won't change
	memset( &mapthead, 0, sizeof( MAPTHEAD_STRUCT ) );

	mapthead.maptype = 7;	/* bg, fg, and info planes */
	mapthead.tsize = 2;	/* 16x16 tiles */

	mapthead.RLEWtag = maphead.RLEWtag;

	switch( exe_info[v].episode ) {
		case 4: mapthead.NumIconRows = 5; break;
		case 5: mapthead.NumIconRows = 7; break;
		case 6: mapthead.NumIconRows = 7; break;
	}

	// Now to decarmack the maps into maptemp
	printf( "Extracting level data...\n" );
	do_maps( v );

	// Now write the mapthead struct
	sprintf( fname, "mapthead.ck%c", ckver );
	fo = fopen( fname, "wb" );
	if( fo ) {
		fwrite( &mapthead, sizeof( MAPTHEAD_STRUCT ), 1, fo );
		fclose( fo );
	} else {
		printf( "Error: can't create '%s'!\n", fname );
	}

	return 0;
}

int process_keen_exe( char *filename ) {
	FILE *f;
	long imglen, headerlen;
	int i, ret = -1;

	printf( "\nProcessing %s...\n", filename );

	// Open the file
	f = fopen( filename, "rb" );
	if( f ) {
		// Try and identify it
		get_exe_image_size( f, &imglen, &headerlen );
		for( i = 0; i < NUM_VERSIONS; i++ ) {
			if( imglen == exe_info[i].imglen )
				break;
		}

		if( i < NUM_VERSIONS ) {
			// We have identified it!
			printf( "Found Keen %d version %d.%d\n", exe_info[i].episode,
				exe_info[i].version / 100, exe_info[i].version % 100 );

			ret = extract_data( f, i, headerlen );
		} else {
			// Couldn't identify it
			printf( "Couldn't identify %s! Perhaps you need to "
					"decompress it with UNP or UNLZEXE?\n", filename );
		}
		fclose( f );
	} else {
		printf( "Error: can't open %s!\n", filename );
	}

	return ret;
}



int main(int argc, char *argv[]) {
	struct ffblk fb;
	int more;

	setbuf( stdout, NULL );

	printf("TEDSETUP v1.02\tCopyright (c)2003-4 Andrew Durdin.\n\n"
		   "Use the switch /maphead:filename to get maphead data\n"
		   "from a file instead of the .exe file (so that you can"
		   "rescue a TED5 project.\n\n");

	if(argc > 1)
	{
		if(strncmp(argv[1], "/maphead:", 9) == 0)
		{
			externalmaphead = 1;
			strncpy(mapheadname, argv[1]+9, 31);
			mapheadname[31] = 0;
		}
	}

	more = !findfirst( "k*.exe", &fb, 0 );

	while( more ) {
		process_keen_exe( fb.ff_name );

		more = !findnext( &fb );
	}

	printf("All Done!\n");
	return 0;
}
